@@ -1,4 +1,5 @@
 // SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
+// SPDX-FileCopyrightText: 2023 ivan tkachenko <me@ratijas.tk>
 // SPDX-FileCopyrightText: 2023 Harald Sitter <sitter@kde.org>
 
 #include <QAbstractItemModel>
@@ -180,7 +181,7 @@ private Q_SLOTS:
         model.load();
         bool containsNetwork = false;
         bool containsXdgDownload = false;
-        for (auto i = 0; i <= model.rowCount(); ++i) {
+        for (auto i = 0; i < model.rowCount(); ++i) {
             const QString name = model.data(model.index(i, 0), FlatpakPermissionModel::Name).toString();
             if (name == "network") {
                 containsNetwork = true;
@@ -188,15 +189,75 @@ private Q_SLOTS:
 
             if (name == "xdg-download") {
                 containsXdgDownload = true;
+                QVERIFY(!model.data(model.index(i, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
                 QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled), true);
-                QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::EffectiveValue), i18n("read/write"));
+                QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::EffectiveValue).value<FlatpakFilesystemsEntry::AccessMode>(),
+                         FlatpakFilesystemsEntry::AccessMode::ReadWrite);
             }
         }
 
         QVERIFY(containsNetwork);
         QVERIFY(containsXdgDownload);
-        QVERIFY(model.permExists("network"));
-        QVERIFY(!model.permExists("yolo-foobar"));
+        QVERIFY(model.permissionExists(FlatpakPermissionsSectionType::Basic, QLatin1String("network")));
+        QVERIFY(!model.permissionExists(FlatpakPermissionsSectionType::Basic, QLatin1String("yolo-foobar")));
+    }
+
+    void testLoadSameSimpleOverride()
+    {
+        // Test that values that are for some reason enabled and present in
+        // both defaults and overrides are parsed correctly.
+        FlatpakReferencesModel referencesModel;
+        QFile metadataFile(QFINDTESTDATA("fixtures/metadata/com.example.same.override"));
+        QVERIFY(metadataFile.open(QFile::ReadOnly));
+        FlatpakReference reference(&referencesModel,
+                                   "com.example.same.override",
+                                   "x86_64",
+                                   "stable",
+                                   "0.0.24",
+                                   "Test Same Simple Override",
+                                   QFINDTESTDATA("fixtures/overrides/"),
+                                   QUrl(),
+                                   metadataFile.readAll());
+        FlatpakPermissionModel model;
+        model.setReference(&reference);
+        model.setShowAdvanced(true);
+        model.load();
+
+        QStringList found;
+
+        for (auto i = 0; i < model.rowCount(); ++i) {
+            const auto name = model.data(model.index(i, 0), FlatpakPermissionModel::Name).toString();
+
+            const auto isDefaultEnabled = model.data(model.index(i, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool();
+            const auto isEffectiveEnabled = model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled).toBool();
+
+            if (name == QLatin1String("ipc")) {
+                QVERIFY(isDefaultEnabled);
+                QVERIFY(isEffectiveEnabled);
+            } else if (name == QLatin1String("pulseaudio")) {
+                QVERIFY(isDefaultEnabled);
+                QVERIFY(isEffectiveEnabled);
+            } else if (name == QLatin1String("dri")) {
+                QVERIFY(!isDefaultEnabled);
+                QVERIFY(!isEffectiveEnabled);
+            } else if (name == QLatin1String("bluetooth")) {
+                QVERIFY(!isDefaultEnabled);
+                QVERIFY(isEffectiveEnabled);
+            } else if (name == QLatin1String("xdg-download")) {
+                QVERIFY(isDefaultEnabled);
+                QVERIFY(isEffectiveEnabled);
+            } else {
+                continue;
+            }
+
+            found.append(name);
+        }
+
+        QVERIFY(found.contains(QLatin1String("ipc")));
+        QVERIFY(found.contains(QLatin1String("pulseaudio")));
+        QVERIFY(found.contains(QLatin1String("dri")));
+        QVERIFY(found.contains(QLatin1String("bluetooth")));
+        QVERIFY(found.contains(QLatin1String("xdg-download")));
     }
 
     void testDefaultFilesystemsGoFirst()
@@ -221,7 +282,7 @@ private Q_SLOTS:
         model.setReference(&reference);
         model.load();
         QStringList filesystems;
-        for (auto i = 0; i <= model.rowCount(); ++i) {
+        for (auto i = 0; i < model.rowCount(); ++i) {
             const QString name = model.data(model.index(i, 0), FlatpakPermissionModel::Name).toString();
             // collect all filesystems
             const auto section = model.data(model.index(i, 0), FlatpakPermissionModel::Section).value<FlatpakPermissionsSectionType::Type>();
@@ -273,7 +334,7 @@ private Q_SLOTS:
         const auto service2 = QLatin1String("com.example.service2");
         int indexOfService2 = -1;
 
-        for (auto i = 0; i <= model.rowCount(); ++i) {
+        for (auto i = 0; i < model.rowCount(); ++i) {
             const QString name = model.data(model.index(i, 0), FlatpakPermissionModel::Name).toString();
             const auto section = model.data(model.index(i, 0), FlatpakPermissionModel::Section).value<FlatpakPermissionsSectionType::Type>();
             if (section == FlatpakPermissionsSectionType::SessionBus) {
@@ -295,15 +356,15 @@ private Q_SLOTS:
         QCOMPARE(name0, FlatpakPolicy::FLATPAK_POLICY_NONE);
         QCOMPARE(name1, FlatpakPolicy::FLATPAK_POLICY_NONE);
         QCOMPARE(name2, FlatpakPolicy::FLATPAK_POLICY_SEE);
-        const auto isDefaultEnabled0 = model.data(model.index(indexOfService0, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool();
-        const auto isDefaultEnabled1 = model.data(model.index(indexOfService1, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool();
-        const auto isDefaultEnabled2 = model.data(model.index(indexOfService2, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool();
-        QVERIFY(isDefaultEnabled0);
-        QVERIFY(!isDefaultEnabled1);
-        QVERIFY(!isDefaultEnabled2);
+        QVERIFY(model.data(model.index(indexOfService0, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool());
+        QVERIFY(!model.data(model.index(indexOfService1, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool());
+        QVERIFY(!model.data(model.index(indexOfService2, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool());
+        QVERIFY(!model.data(model.index(indexOfService0, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
+        QVERIFY(model.data(model.index(indexOfService1, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
+        QVERIFY(model.data(model.index(indexOfService2, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
 
         const auto setAndCheckBus = [&](const QString &value, FlatpakPolicy newPolicy) {
-            model.editPerm(indexOfService2, newPolicy);
+            model.setPermissionValueAtRow(indexOfService2, newPolicy);
             const auto name2i18nValue = model.data(model.index(indexOfService2, 0), FlatpakPermissionModel::EffectiveValue).value<FlatpakPolicy>();
             QCOMPARE(name2i18nValue, newPolicy);
 
@@ -340,13 +401,13 @@ private Q_SLOTS:
         QVERIFY(isEffectiveEnabled(indexOfService2));
 
         // Toggling non-default bus entry should disable it (i.e. mark for deletion)
-        model.togglePermissionAtIndex(indexOfService2);
+        model.togglePermissionAtRow(indexOfService2);
         QVERIFY(!isEffectiveEnabled(indexOfService2));
         // Reloading data should re-enable it again.
         model.load();
         QVERIFY(isEffectiveEnabled(indexOfService2));
         // Disabling non-default bus entry and saving it should remove it from override file.
-        model.togglePermissionAtIndex(indexOfService2);
+        model.togglePermissionAtRow(indexOfService2);
         QVERIFY(!isEffectiveEnabled(indexOfService2));
         model.save();
         {
@@ -355,7 +416,7 @@ private Q_SLOTS:
             QVERIFY(!group.hasKey(service2));
         }
         // Re-enabling non-default bus entry and saving it should add it back to override file.
-        model.togglePermissionAtIndex(indexOfService2);
+        model.togglePermissionAtRow(indexOfService2);
         QVERIFY(isEffectiveEnabled(indexOfService2));
         model.save();
         {
@@ -394,7 +455,7 @@ private Q_SLOTS:
         const auto system = QLatin1String("org.freedesktop.timedate1");
         int indexOfSystem = -1;
 
-        for (auto i = 0; i <= model.rowCount(); ++i) {
+        for (auto i = 0; i < model.rowCount(); ++i) {
             const auto name = model.data(model.index(i, 0), FlatpakPermissionModel::Name).toString();
             const auto section = model.data(model.index(i, 0), FlatpakPermissionModel::Section).value<FlatpakPermissionsSectionType::Type>();
             if (section == FlatpakPermissionsSectionType::SessionBus) {
@@ -456,45 +517,52 @@ private Q_SLOTS:
         const auto envName = QLatin1String("SOME_ENV");
         const auto envValue = QLatin1String("abc123");
 
-        model.addUserEnteredPermission(FlatpakPermissionsSectionType::Filesystems, filesystem, i18n("create"));
+        model.addUserEnteredPermission(FlatpakPermissionsSectionType::Filesystems,
+                                       filesystem,
+                                       QVariant::fromValue(FlatpakFilesystemsEntry::AccessMode::Create));
         model.addUserEnteredPermission(FlatpakPermissionsSectionType::SessionBus, session, FlatpakPolicy::FLATPAK_POLICY_TALK);
         // Try int cast to make sure QML/JS works fine too.
-        model.addUserEnteredPermission(FlatpakPermissionsSectionType::SystemBus, system, static_cast<int>(FlatpakPolicy::FLATPAK_POLICY_SEE));
+        model.addUserEnteredPermission(FlatpakPermissionsSectionType::SystemBus, system, FlatpakPolicy::FLATPAK_POLICY_SEE);
         model.addUserEnteredPermission(FlatpakPermissionsSectionType::Environment, envName, envValue);
 
-        for (auto i = 0; i <= model.rowCount(); ++i) {
+        for (auto i = 0; i < model.rowCount(); ++i) {
             const QString name = model.data(model.index(i, 0), FlatpakPermissionModel::Name).toString();
             if (name == "host") {
                 const auto metaEnum = QMetaEnum::fromType<FlatpakPermissionModel::Roles>();
-                for (auto j = 0; j <= metaEnum.keyCount(); ++j) { // purely for debugging purposes
+                for (auto j = 0; j < metaEnum.keyCount(); ++j) { // purely for debugging purposes
                     qDebug() << metaEnum.key(j) << model.data(model.index(i, 0), metaEnum.value(j));
                 }
 
                 QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled), false);
-                model.togglePermissionAtIndex(i);
+                QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
+                model.togglePermissionAtRow(i);
                 QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled), true);
             }
 
             if (name == "org.kde.StatusNotifierWatcher") {
-                model.editPerm(i, FlatpakPolicy::FLATPAK_POLICY_OWN);
+                model.setPermissionValueAtRow(i, FlatpakPolicy::FLATPAK_POLICY_OWN);
             }
 
             if (name == "host-os") {
                 // Make sure the config manipulation works across multiple changes
-                model.editPerm(i, i18n("read-only"));
-                model.editPerm(i, i18n("read-write"));
-                model.editPerm(i, i18n("create"));
-                model.editPerm(i, i18n("read-only"));
+                model.setPermissionValueAtRow(i, QVariant::fromValue(FlatpakFilesystemsEntry::AccessMode::ReadOnly));
+                model.setPermissionValueAtRow(i, QVariant::fromValue(FlatpakFilesystemsEntry::AccessMode::ReadWrite));
+                model.setPermissionValueAtRow(i, QVariant::fromValue(FlatpakFilesystemsEntry::AccessMode::Create));
+                model.setPermissionValueAtRow(i, QVariant::fromValue(FlatpakFilesystemsEntry::AccessMode::ReadOnly));
             }
 
             if (name == filesystem) {
+                QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
                 QVERIFY(!model.data(model.index(i, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool());
                 QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled).toBool());
-                QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::DefaultValue).toString(), i18n("create"));
-                QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::EffectiveValue).toString(), i18n("create"));
+                QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::DefaultValue).value<FlatpakFilesystemsEntry::AccessMode>(),
+                         FlatpakFilesystemsEntry::AccessMode::Create);
+                QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::EffectiveValue).value<FlatpakFilesystemsEntry::AccessMode>(),
+                         FlatpakFilesystemsEntry::AccessMode::Create);
             }
 
             if (name == session) {
+                QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
                 QVERIFY(!model.data(model.index(i, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool());
                 QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled).toBool());
                 QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::DefaultValue).value<FlatpakPolicy>(), FlatpakPolicy::FLATPAK_POLICY_TALK);
@@ -502,6 +570,7 @@ private Q_SLOTS:
             }
 
             if (name == system) {
+                QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
                 QVERIFY(!model.data(model.index(i, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool());
                 QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled).toBool());
                 QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::DefaultValue).value<FlatpakPolicy>(), FlatpakPolicy::FLATPAK_POLICY_SEE);
@@ -509,6 +578,7 @@ private Q_SLOTS:
             }
 
             if (name == envName) {
+                QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::CanBeDisabled).toBool());
                 QVERIFY(!model.data(model.index(i, 0), FlatpakPermissionModel::IsDefaultEnabled).toBool());
                 QVERIFY(model.data(model.index(i, 0), FlatpakPermissionModel::IsEffectiveEnabled).toBool());
                 QCOMPARE(model.data(model.index(i, 0), FlatpakPermissionModel::DefaultValue).toString(), envValue);
@@ -530,8 +600,7 @@ private Q_SLOTS:
     void testValuesModelForFilesystemsSection()
     {
         const auto values = FlatpakPermissionModel::valuesModelForFilesystemsSection();
-        QCOMPARE(values->rowCount(QModelIndex()), 3);
-        QEXPECT_FAIL("", "Filesystems logic is not ready to use 'OFF' value yet.", Continue);
+        QCOMPARE(values->rowCount(QModelIndex()), 4);
         QVERIFY(modelContains(values, i18n("OFF"))); // Filesystems logic is not ready to use this value yet.
         QVERIFY(modelContains(values, i18n("read/write")));
         QVERIFY(modelContains(values, i18n("read-only")));
@@ -561,6 +630,120 @@ private Q_SLOTS:
         actual = FlatpakPermissionModel::valuesModelForSectionType(FlatpakPermissionsSectionType::SystemBus);
         QCOMPARE(actual, expected);
     }
+
+    void testUnparsableFilesystems()
+    {
+        FlatpakReferencesModel referencesModel;
+        QFile metadataFile(QFINDTESTDATA("fixtures/metadata/com.example.unparsable.filesystems"));
+        QVERIFY(metadataFile.open(QFile::ReadOnly));
+        FlatpakReference reference(&referencesModel,
+                                   "com.example.unparsable.filesystems",
+                                   "x86_64",
+                                   "stable",
+                                   "0.0.24",
+                                   "Unparsable Filesystems",
+                                   QFINDTESTDATA("fixtures/overrides/"),
+                                   QUrl(),
+                                   metadataFile.readAll());
+        FlatpakPermissionModel model;
+        model.setReference(&reference);
+        model.load();
+        model.setShowAdvanced(true);
+
+        const auto filesystem = QLatin1String("xdg-data/path");
+        model.addUserEnteredPermission(FlatpakPermissionsSectionType::Filesystems,
+                                       filesystem,
+                                       QVariant::fromValue(FlatpakFilesystemsEntry::AccessMode::ReadOnly));
+        model.save();
+
+        const KConfig actual(QFINDTESTDATA("fixtures/overrides/com.example.unparsable.filesystems"));
+        const KConfig expected(QFINDTESTDATA("fixtures/overrides.out/com.example.unparsable.filesystems"));
+        QVERIFY(operatorFlatpakConfigEquals(actual, expected));
+    }
+
+    void testDefaultHomeFilesystem()
+    {
+        FlatpakReferencesModel referencesModel;
+        QFile metadataFile(QFINDTESTDATA("fixtures/metadata/com.example.home.filesystems"));
+        QVERIFY(metadataFile.open(QFile::ReadOnly));
+        FlatpakReference reference(&referencesModel,
+                                   "com.example.home.filesystems",
+                                   "x86_64",
+                                   "stable",
+                                   "0.0.24",
+                                   "Default Home Filesystems",
+                                   QFINDTESTDATA("fixtures/overrides/"),
+                                   QUrl(),
+                                   metadataFile.readAll());
+        FlatpakPermissionModel model;
+        model.setReference(&reference);
+        model.load();
+
+        int dConfIndex = -1;
+        int homeIndex = -1;
+
+        for (int i = 0; i < model.rowCount(); i++) {
+            const auto section = model.data(model.index(i), FlatpakPermissionModel::Section).value<FlatpakPermissionsSectionType::Type>();
+            const auto name = model.data(model.index(i), FlatpakPermissionModel::Name).toString();
+
+            if (section == FlatpakPermissionsSectionType::Filesystems) {
+                const auto value = model.data(model.index(i), FlatpakPermissionModel::DefaultValue).value<FlatpakFilesystemsEntry::AccessMode>();
+
+                if (name.contains(QLatin1String("dconf"))) {
+                    dConfIndex = i;
+                    QCOMPARE(value, FlatpakFilesystemsEntry::AccessMode::ReadWrite);
+                } else if (name == QLatin1String("home")) {
+                    homeIndex = i;
+                    QCOMPARE(value, FlatpakFilesystemsEntry::AccessMode::ReadOnly);
+                }
+            }
+        }
+
+        QVERIFY(dConfIndex != -1);
+        QVERIFY(homeIndex != -1);
+    }
+
+    void testValidNames()
+    {
+        FlatpakReferencesModel referencesModel;
+        QFile metadataFile(QFINDTESTDATA("fixtures/metadata/com.example.valid.names"));
+        QVERIFY(metadataFile.open(QFile::ReadOnly));
+        FlatpakReference reference(&referencesModel,
+                                   "com.example.valid.names",
+                                   "x86_64",
+                                   "stable",
+                                   "1.0",
+                                   "Valid Names",
+                                   QFINDTESTDATA("fixtures/overrides/"),
+                                   QUrl(),
+                                   metadataFile.readAll());
+        FlatpakPermissionModel model;
+        model.setReference(&reference);
+        model.load();
+        model.setShowAdvanced(true);
+
+        QVERIFY(!model.isFilesystemNameValid(QString()));
+        QVERIFY(model.permissionExists(FlatpakPermissionsSectionType::Filesystems, QLatin1String("~/path")));
+        QVERIFY(model.isFilesystemNameValid(QLatin1String("~/valid/path")));
+        QVERIFY(model.isFilesystemNameValid(QLatin1String("~/other")));
+
+        QVERIFY(!model.isDBusServiceNameValid(QString()));
+        QVERIFY(!model.isDBusServiceNameValid(QLatin1String("com")));
+        QVERIFY(!model.isDBusServiceNameValid(QLatin1String("#$%")));
+        QVERIFY(model.permissionExists(FlatpakPermissionsSectionType::SessionBus, QLatin1String("com.example.session")));
+        QVERIFY(model.permissionExists(FlatpakPermissionsSectionType::SystemBus, QLatin1String("com.example.system")));
+        QVERIFY(model.isDBusServiceNameValid(QLatin1String("com.example.session")));
+        QVERIFY(model.isDBusServiceNameValid(QLatin1String("com.example.system")));
+        QVERIFY(model.isDBusServiceNameValid(QLatin1String("com.example.session2")));
+        QVERIFY(model.isDBusServiceNameValid(QLatin1String("com.example.*")));
+        QVERIFY(!model.isDBusServiceNameValid(QLatin1String("com.example.")));
+
+        QVERIFY(!model.isEnvironmentVariableNameValid(QString()));
+        QVERIFY(!model.isEnvironmentVariableNameValid(QLatin1String("ABC=DEF")));
+        QVERIFY(model.permissionExists(FlatpakPermissionsSectionType::Environment, QLatin1String("EXAMPLE_NAME")));
+        QVERIFY(model.isEnvironmentVariableNameValid(QLatin1String("EXAMPLE_NAME")));
+        QVERIFY(model.isEnvironmentVariableNameValid(QLatin1String("whatever")));
+    }
 };
 
 QTEST_MAIN(FlatpakPermissionModelTest)
