@@ -1,5 +1,6 @@
 /**
  * SPDX-FileCopyrightText: 2022 Suhaas Joshi <joshiesuhaas0@gmail.com>
+ * SPDX-FileCopyrightText: 2023 ivan tkachenko <me@ratijas.tk>
  * SPDX-License-Identifier: GPL-2.0-or-later
  */
 
@@ -8,14 +9,18 @@
 #include <QDir>
 #include <QFileInfo>
 #include <QStandardPaths>
+#include <QUrl>
+
+#include <gio/gio.h>
 
 namespace FlatpakHelper
 {
-QString permDataFilePath()
+
+QString permissionsDataDirectory()
 {
-    QString userPath = QString::fromStdString(qgetenv("FLATPAK_USER_DIR").toStdString());
+    QString userPath = qEnvironmentVariable("FLATPAK_USER_DIR");
     if (userPath.isEmpty()) {
-        userPath = QString::fromStdString(qgetenv("HOST_XDG_DATA_HOME").toStdString());
+        userPath = qEnvironmentVariable("HOST_XDG_DATA_HOME");
         if (userPath.isEmpty()) {
             userPath = QStandardPaths::writableLocation(QStandardPaths::GenericDataLocation);
         }
@@ -25,7 +30,7 @@ QString permDataFilePath()
     return userPath;
 }
 
-QString iconPath(const QString &name, const QString &id, const QString &appBasePath)
+QUrl iconSourceUrl(const QString &displayName, const QString &flatpakName, const QString &appBasePath)
 {
     QString dirPath = appBasePath + QStringLiteral("/files/share/icons/hicolor/");
     QDir dir(dirPath);
@@ -39,24 +44,37 @@ QString iconPath(const QString &name, const QString &id, const QString &appBaseP
     } else if (!dir.isEmpty()) {
         nextDir = dir.entryList().at(0);
     } else {
-        return QString();
+        return QUrl();
     }
     dir.cd(nextDir + QStringLiteral("/apps"));
 
-    QString file = id + QStringLiteral(".png");
+    QString file = flatpakName + QStringLiteral(".png");
     if (!dir.exists(file)) {
-        file = id + QStringLiteral(".svg");
+        file = flatpakName + QStringLiteral(".svg");
         if (!dir.exists(file)) {
-            file = name.toLower() + QStringLiteral(".png");
+            file = displayName.toLower() + QStringLiteral(".png");
             if (!dir.exists(file)) {
-                file = name.toLower() + QStringLiteral(".svg");
+                file = displayName.toLower() + QStringLiteral(".svg");
                 if (!dir.exists(file)) {
-                    return id + QStringLiteral(".png");
+                    return QUrl::fromLocalFile(flatpakName + QStringLiteral(".png"));
                 }
             }
         }
     }
-    return dir.absoluteFilePath(file);
+    return QUrl::fromLocalFile(dir.absoluteFilePath(file));
 }
 
+bool verifyDBusName(QStringView name)
+{
+    if (name.endsWith(QLatin1String(".*"))) {
+        name.chop(2);
+        name = name.mid(-2);
+    }
+
+    const auto ownedName = name.toString();
+    const auto arrayName = ownedName.toUtf8();
+    const auto cName = arrayName.constData();
+
+    return g_dbus_is_name(cName) && !g_dbus_is_unique_name(cName);
+}
 }
