@@ -1,5 +1,6 @@
 /**
  * SPDX-FileCopyrightText: 2022 Suhaas Joshi <joshiesuhaas0@gmail.com>
+ * SPDX-FileCopyrightText: 2023 ivan tkachenko <me@ratijas.tk>
  * SPDX-License-Identifier: GPL-2.0-or-later
  */
 
@@ -8,6 +9,7 @@
 #include "flatpakcommon.h"
 #include "flatpakreference.h"
 
+#include <KConfigGroup>
 #include <QAbstractItemModel>
 #include <QAbstractListModel>
 #include <QPointer>
@@ -39,6 +41,64 @@ public:
     Q_ENUM(Type)
 };
 
+/**
+ * Boolean entry of one of the `shared=`, `sockets=`, `devices=` or `features=`
+ * lists of the `[Context]` group.
+ *
+ * Such entries are either ON or OFF. They are turned OFF if and only if their
+ * string representation is prefixed with a `!` bang character.
+ */
+class FlatpakSimpleEntry
+{
+    Q_GADGET
+public:
+    // Default constructor is required for meta-type registration.
+    /** Default constructor. Creates an invalid entry. */
+    FlatpakSimpleEntry() = default;
+
+    /**
+     * Construct new entry directly from raw unvalidated data.
+     */
+    explicit FlatpakSimpleEntry(const QString &name, bool enabled = true);
+
+    /**
+     * Parse list entry into data structure. Might fail for various reasons,
+     * such as ill-formed access mode, so returns an optional value.
+     */
+    static std::optional<FlatpakSimpleEntry> parse(QStringView entry);
+
+    /**
+     * Parse list of entries. Returns list of successfully parsed valid entries,
+     * as well as a list of the rest unparsable strings.
+     */
+    static std::pair<QStringList, QList<FlatpakSimpleEntry>> getCategory(const KConfigGroup &group, const QString &category);
+
+    /**
+     * Parse list of entries. Returns only list of successfully parsed valid entries.
+     */
+    static QList<FlatpakSimpleEntry> getCategorySkippingInvalidEntries(const KConfigGroup &group, const QString &category);
+
+    static std::optional<bool> isEnabled(const QList<FlatpakSimpleEntry> &entries, const QString &name);
+
+    /**
+     * Format this entry into string. Opposite of parse().
+     */
+    QString format() const;
+
+    bool isEnabled() const;
+    void setEnabled(bool enabled);
+
+    const QString &name() const;
+
+    // TODO C++20: use `= default` implementation.
+    bool operator==(const FlatpakSimpleEntry &other) const;
+    bool operator!=(const FlatpakSimpleEntry &other) const;
+
+private:
+    QString m_name;
+    bool m_enabled = true;
+};
+
 /**
  * Entry of the `filesystems=` list.
  *
@@ -50,6 +110,7 @@ public:
  */
 class FlatpakFilesystemsEntry
 {
+    Q_GADGET
 public:
     enum class AccessMode {
         /** Make the given directory available read-only. */
@@ -61,6 +122,7 @@ public:
         /** Don't expose filesystem to app. */
         Deny,
     };
+    Q_ENUM(AccessMode)
 
     enum class PathMode {
         Required,
@@ -123,6 +185,22 @@ public:
      */
     static std::optional<FlatpakFilesystemsEntry> parse(QStringView entry);
 
+    /**
+     * Parse name of a list entry into data structure. Access mode is passed
+     * separately and not expected to be a part of the name. Might fail for
+     * various reasons, so returns an optional value.
+     *
+     * See also: name()
+     */
+    static std::optional<FlatpakFilesystemsEntry> parse(QStringView name, AccessMode accessMode);
+
+    /**
+     * Formatted prefix and path only, without access mode prefixes or suffixes. Suitable for comparison.
+     *
+     * See also: parse(QStringView, AccessMode)
+     */
+    QString name() const;
+
     /**
      * Format this entry into string. Opposite of parse(). Omits default `:rw` suffix.
      */
@@ -143,8 +221,6 @@ private:
     QString m_path;
 };
 
-Q_DECLARE_METATYPE(FlatpakFilesystemsEntry)
-
 class PolicyChoicesModel : public QAbstractListModel
 {
     Q_OBJECT
@@ -164,7 +240,7 @@ protected:
         QString display;
     };
 
-    PolicyChoicesModel(QVector<Entry> &&policies, QObject *parent = nullptr);
+    explicit PolicyChoicesModel(QVector<Entry> &&policies, QObject *parent = nullptr);
 
 private:
     const QVector<Entry> m_policies;
@@ -174,14 +250,14 @@ class FilesystemChoicesModel : public PolicyChoicesModel
 {
     Q_OBJECT
 public:
-    FilesystemChoicesModel(QObject *parent = nullptr);
+    explicit FilesystemChoicesModel(QObject *parent = nullptr);
 };
 
 class DBusPolicyChoicesModel : public PolicyChoicesModel
 {
     Q_OBJECT
 public:
-    DBusPolicyChoicesModel(QObject *parent = nullptr);
+    explicit DBusPolicyChoicesModel(QObject *parent = nullptr);
 };
 
 /**
@@ -241,10 +317,8 @@ public:
          * Predefined resources come with translated description.
          */
         // TODO: Instead of unchecking there should be more obvious UI. For Bus
-        // type, there's a "none" value. For environment we should implement
-        // "unset-environment" category. For filesystem paths it is not
-        // specified, so we probably should remove the ability to "uncheck"
-        // them, and show a "Remove" button instead.
+        // type, there's a "none" policy. For environment we should implement
+        // "unset-environment" category.
         BuiltIn,
         /**
          * User-defined permissions are resources that user has manually added
@@ -260,7 +334,7 @@ public:
         Dummy
     };
 
-    using Variant = std::variant<QString, FlatpakPolicy>;
+    using Variant = std::variant<QString, FlatpakPolicy, FlatpakFilesystemsEntry::AccessMode>;
 
     // Default constructor is required for meta-type registration.
     /** Default constructor. Creates an invalid entry. */
@@ -343,24 +417,24 @@ public:
      * system defaults, it will be removed from user overrides.
      *
      * For user-defined permissions of other ValueType types, disabling them
-     * would mark them for removal.
+     * would mark them for removal but only if they are disabled (not present)
+     * in defaults.
      */
-    // TODO: What should it do for built-in permissions of other ValueType
-    // types? It doesn't make much sense.
+    bool canBeDisabled() const;
     bool isEffectiveEnabled() const;
     void setEffectiveEnabled(bool enabled);
 
     /**
-     * System default value for this permission. It can not be modified.
+     * System default value for this permission. It can only be modified for
+     * entries which are disabled (not present) by default.
      *
      * Applicable for any permissions other than ValueType::Simple.
      */
     const Variant defaultValue() const;
 
-    // TODO: Remove this method.
-    const Variant overrideValue() const;
+    void setDefaultValue(const Variant &value);
 
-    /** Set user override */
+    /** Set user override. Affects isSaveNeeded() state. */
     void setOverrideValue(const Variant &value);
 
     /**
@@ -372,10 +446,6 @@ public:
     const Variant effectiveValue() const;
     void setEffectiveValue(const Variant &value);
 
-    /** Untranslate value of ValueType::Filesystems permission. */
-    // TODO: Remove this method, store enum variants or otherwise raw untranslated data.
-    QString fsCurrentValue() const;
-
     /** Integration with KCM. */
     bool isSaveNeeded() const;
     bool isDefaults() const;
@@ -435,10 +505,9 @@ public:
         Name,
         Description,
         //
-        IsSimple,
-        IsEnvironment,
         IsNotDummy,
         //
+        CanBeDisabled,
         IsDefaultEnabled,
         IsEffectiveEnabled,
         DefaultValue,
@@ -475,8 +544,6 @@ public:
      *
      * Applicable for ValueType::Filesystems and ValueType::Bus only.
      */
-    // TODO: It should be a model that also contains detailed description
-    // (help text) and untranslated value identifier.
     Q_INVOKABLE static PolicyChoicesModel *valuesModelForSectionType(int /*FlatpakPermissionsSectionType::Type*/ rawSection);
     Q_INVOKABLE static PolicyChoicesModel *valuesModelForFilesystemsSection();
     Q_INVOKABLE static PolicyChoicesModel *valuesModelForBusSections();
@@ -484,9 +551,19 @@ public:
     Q_INVOKABLE static QString sectionHeaderForSectionType(int /*FlatpakPermissionsSectionType::Type*/ rawSection);
     Q_INVOKABLE static QString sectionAddButtonToolTipTextForSectionType(int /*FlatpakPermissionsSectionType::Type*/ rawSection);
 
+    Q_INVOKABLE bool permissionExists(int /*FlatpakPermissionsSectionType::Type*/ rawSection, const QString &name) const;
+    bool permissionExists(FlatpakPermissionsSectionType::Type section, const QString &name) const;
+
+    /**
+     * Validators to check that names comply with section-specific rules.
+     */
+    Q_INVOKABLE static bool isFilesystemNameValid(const QString &name);
+    Q_INVOKABLE static bool isDBusServiceNameValid(const QString &name);
+    Q_INVOKABLE static bool isEnvironmentVariableNameValid(const QString &name);
+
 public Q_SLOTS:
-    void togglePermissionAtIndex(int index);
-    void editPerm(int index, const QVariant &newValue);
+    void togglePermissionAtRow(int row);
+    void setPermissionValueAtRow(int row, const QVariant &newValue);
     void addUserEnteredPermission(int /*FlatpakPermissionsSectionType::Type*/ rawSection, const QString &name, const QVariant &value);
 
 Q_SIGNALS:
@@ -494,27 +571,15 @@ Q_SIGNALS:
     void showAdvancedChanged();
 
 private:
-    void addPermission(const FlatpakPermission &permission, bool shouldBeOn);
-    void removePermission(const FlatpakPermission &permission, bool isGranted);
+    int findIndexToInsertRowAndRemoveDummyRowIfNeeded(FlatpakPermissionsSectionType::Type section, bool emitModelSignals);
 
-    void editFilesystemsPermissions(FlatpakPermission &permission, const QString &newValue);
-
-    void addBusPermissions(const FlatpakPermission &permission);
-    void removeBusPermission(const FlatpakPermission &permission);
-    void editBusPermissions(FlatpakPermission &permission, FlatpakPolicy newValue);
-
-    void addEnvPermission(const FlatpakPermission &permission);
-    void removeEnvPermission(const FlatpakPermission &permission);
-    void editEnvPermission(FlatpakPermission &permission, const QString &newValue);
-
-    bool permExists(const QString &name) const;
-    int permIndex(const QString &category);
-
-    void readFromFile();
-    void writeToFile();
+    void writeToFile() const;
+    void writeToKConfig(KConfig &config) const;
 
     QVector<FlatpakPermission> m_permissions;
+    QHash<QString, QStringList> m_unparsableEntriesByCategory;
     QPointer<FlatpakReference> m_reference;
-    QString m_overridesData;
     bool m_showAdvanced;
+
+    friend class FlatpakPermissionModelTest;
 };
